import json
import logging
import os
import warnings
from json.decoder import JSONDecodeError
from google import auth
from google.oauth2 import service_account
from octue.exceptions import InvalidInputException


logger = logging.getLogger(__name__)

DEFAULT_ENVIRONMENT_VARIABLE_NAME = "GOOGLE_APPLICATION_CREDENTIALS"


class GCPCredentialsManager:
    """A credentials manager for Google Cloud Platform (GCP) that takes a path to a service account JSON file, or a
    JSON string of the contents of such a service account file, from the given environment variable and instantiates
    a Google Cloud credentials object.

    :param str environment_variable_name:
    :return None:
    """

    def __init__(self, environment_variable_name=DEFAULT_ENVIRONMENT_VARIABLE_NAME):

        if environment_variable_name is None:
            raise InvalidInputException(
                "Cannot accept `None` as an environment variable name. Either specify the name as a string or use the default value"
            )

        self.environment_variable_name = environment_variable_name

        # Note: We use '' rather than None to allow us to patch the value in testing (you cannot correctly unset variables in a patch of os.environ)
        self.environment_variable_value = os.getenv(environment_variable_name, "")

        if self.environment_variable_value == "":
            if self.environment_variable_name != DEFAULT_ENVIRONMENT_VARIABLE_NAME:
                raise InvalidInputException(
                    f"Specified use of credentials environment variable {environment_variable_name} but its value is unset"
                )

        logger.debug("Initialised CredentialsManager with %s", environment_variable_name)

    @property
    def using_application_default_credentials(self):
        """A boolean determining whether ADCs are used or not"""
        return self.environment_variable_value == ""

    def get_credentials(self, as_dict=False):
        """Get the Google OAUTH2 service account credentials.

        :param bool as_dict: if `True`, get the credentials as a dictionary
        :return dict|google.auth.service_account.Credentials|None:
        """
        if as_dict:
            warnings.warn(
                message=(
                    "Requesting credentials as a dictionary is deprecated "
                    "to enable uniform treatment between specified "
                    "GOOGLE_APPLICATION_CREDENTIALS and use of Google's "
                    "Application Default Credentials."
                ),
                category=DeprecationWarning,
            )

        if self.using_application_default_credentials:
            logger.debug("Using application default credentials")
            creds, project_id = auth.default()
            # FFS google why can't you bloody make it simple
            creds.project_id = project_id

        elif os.path.exists(self.environment_variable_value):
            logger.debug("Using credentials from file %s", self.environment_variable_value)
            creds = self._get_credentials_from_file(self.environment_variable_value, as_dict=as_dict)

        else:
            warnings.warn(
                message=(
                    "Placing service account credentials JSON directly in an "
                    "environment variable will be deprecated soon. Set "
                    "GOOGLE_APPLICATION_CREDENTIALS to a file path, or leave "
                    "unset to use Application Default Credentials. "
                    "See https://medium.com/datamindedbe/application-default-credentials-477879e31cb5 "
                    "for a helpful overview of google credentials"
                ),
                category=DeprecationWarning,
            )

            logger.debug("Using credentials from environment string")
            try:
                creds = self._get_credentials_from_string(self.environment_variable_value, as_dict=as_dict)
            except JSONDecodeError as e:
                raise InvalidInputException(
                    f"The environment variable {self.environment_variable_name} has a value, but is neither a valid file path not valid json. Value is: {self.environment_variable_value}"
                ) from e

        return creds

    def _get_credentials_from_file(self, filename, as_dict=False):
        """Get the credentials from the JSON file whose path is specified in the environment variable's value.

        :param bool as_dict: if `True`, get the credentials as a dictionary
        :return dict|google.auth.service_account.Credentials:
        """
        with open(filename, encoding="utf-8") as f:
            credentials = json.load(f)

        logger.debug("GCP credentials read from file.")

        if as_dict:
            return credentials

        return service_account.Credentials.from_service_account_info(credentials)

    def _get_credentials_from_string(self, value, as_dict=False):
        """Get the credentials directly from the JSON string specified in the environment variable's value.

        :param bool as_dict: if `True`, get the credentials as a dictionary
        :return dict|google.auth.service_account.Credentials:
        """
        credentials = json.loads(value)
        logger.debug("GCP credentials loaded from string.")

        if as_dict:
            return credentials

        return service_account.Credentials.from_service_account_info(credentials)



import concurrent.futures
import json
import os

from octue import definitions
from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.exceptions import InvalidInputException
from octue.migrations.cloud_storage import translate_bucket_name_and_path_in_bucket_to_cloud_path
from octue.mixins import Hashable, Identifiable, Labelable, Pathable, Serialisable, Taggable
from octue.resources.datafile import Datafile
from octue.resources.filter_containers import FilterSet
from octue.resources.label import LabelSet
from octue.resources.tag import TagDict


DATAFILES_DIRECTORY = "datafiles"


class Dataset(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashable):
    """A representation of a dataset, containing files, labels, etc

    This is used to read a list of files (and their associated properties) into octue analysis, or to compile a
    list of output files (results) and their properties that will be sent back to the octue system.

    :param iter(dict|octue.resources.datafile.Datafile) files: the files belonging to the dataset
    :return None:
    """

    _ATTRIBUTES_TO_HASH = ("files",)
    _SERIALISE_FIELDS = "files", "name", "labels", "tags", "id", "path"

    def __init__(self, files=None, name=None, id=None, path=None, path_from=None, tags=None, labels=None, **kwargs):
        super().__init__(name=name, id=id, tags=tags, labels=labels, path=path, path_from=path_from)

        # TODO The decoders aren't being used; utils.decoders.OctueJSONDecoder should be used in twined
        #  so that resources get automatically instantiated.
        #  Add a proper `decoder` argument  to the load_json utility in twined so that datasets, datafiles and manifests
        #  get initialised properly, then remove this hackjob.
        self.files = FilterSet()
        self._cloud_path = None

        for file in files or []:
            if isinstance(file, Datafile):
                self.files.add(file)
            else:
                self.files.add(Datafile.deserialise(file, path_from=self))

        self.__dict__.update(**kwargs)

    def __iter__(self):
        yield from self.files

    def __len__(self):
        return len(self.files)

    @classmethod
    def from_local_directory(cls, path_to_directory, recursive=False, **kwargs):
        """Instantiate a Dataset from the files in the given local directory.

        :param str path_to_directory: path to a local directory
        :param bool recursive: if `True`, include all files in the directory's subdirectories recursively
        :param kwargs: other keyword arguments for the `Dataset` instantiation
        :return Dataset:
        """
        datafiles = FilterSet()

        for level, (directory_path, _, filenames) in enumerate(os.walk(path_to_directory)):
            for filename in filenames:

                if not recursive and level > 0:
                    break

                datafiles.add(Datafile(path=os.path.join(directory_path, filename)))

        return Dataset(path=path_to_directory, files=datafiles, **kwargs)

    @classmethod
    def from_cloud(cls, cloud_path=None, bucket_name=None, path_to_dataset_directory=None, recursive=False):
        """Instantiate a Dataset from Google Cloud storage.

        :param str|None cloud_path: full path to dataset directory in cloud storage (e.g. `gs://bucket_name/path/to/dataset`)
        :param bool recursive: if `True`, include in the dataset all files in the subdirectories recursively contained in the dataset directory
        :return Dataset:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_to_dataset_directory)

        bucket_name = storage.path.split_bucket_name_from_gs_path(cloud_path)[0]

        dataset_metadata = cls._get_dataset_metadata(cloud_path=cloud_path)

        if dataset_metadata:
            return Dataset(
                id=dataset_metadata.get("id"),
                name=dataset_metadata.get("name"),
                path=cloud_path,
                tags=TagDict(dataset_metadata.get("tags", {})),
                labels=LabelSet(dataset_metadata.get("labels", [])),
                files=[Datafile(path=path) for path in dataset_metadata["files"]],
            )

        datafiles = FilterSet(
            Datafile(path=storage.path.generate_gs_path(bucket_name, blob.name))
            for blob in GoogleCloudStorageClient().scandir(cloud_path, recursive=recursive)
        )

        dataset = Dataset(path=cloud_path, files=datafiles)
        dataset._upload_dataset_metadata(cloud_path)
        return dataset

    def to_cloud(self, cloud_path=None, bucket_name=None, output_directory=None):
        """Upload a dataset to the given cloud path.

        :param str|None cloud_path: cloud path to store dataset at (e.g. `gs://bucket_name/path/to/dataset`)
        :return str: cloud path for dataset
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, output_directory)

        files_and_paths = []

        for datafile in self.files:
            datafile_path_relative_to_dataset = datafile.path.split(self.path)[-1].strip(os.path.sep).strip("/")
            files_and_paths.append(
                (
                    datafile,
                    storage.path.join(cloud_path, *datafile_path_relative_to_dataset.split(os.path.sep)),
                )
            )

        def upload(iterable_element):
            """Upload a datafile to the given cloud path.

            :param tuple(octue.resources.datafile.Datafile, str) iterable_element:
            :return None:
            """
            datafile = iterable_element[0]
            cloud_path = iterable_element[1]
            datafile.to_cloud(cloud_path=cloud_path)

        # Use multiple threads to significantly speed up file uploads by reducing latency.
        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(upload, files_and_paths)

        self._upload_dataset_metadata(cloud_path)
        self._cloud_path = cloud_path
        return cloud_path

    @property
    def name(self):
        """Get the name of the dataset

        :return str:
        """
        return self._name or os.path.split(os.path.abspath(os.path.split(self.path)[-1]))[-1]

    @property
    def cloud_path(self):
        """Get the cloud path of the dataset.

        :return str|None:
        """
        return self._cloud_path

    @cloud_path.setter
    def cloud_path(self, path):
        """Set the cloud path of the dataset.

        :param str|None path:
        :return None:
        """
        if path is None:
            self._cloud_path = None
            return

        self.to_cloud(cloud_path=path)

    @property
    def all_files_are_in_cloud(self):
        """Do all the files of the dataset exist in the cloud?

        :return bool:
        """
        return all(file.exists_in_cloud for file in self.files)

    def add(self, *args, **kwargs):
        """Add a data/results file to the manifest.

        Usage:
            my_file = octue.DataFile(...)
            my_manifest.add(my_file)

            # or more simply
            my_manifest.add(**{...}) which implicitly creates the datafile from the starred list of input arguments
        """
        if len(args) > 1:
            # Recurse to allow addition of many files at once
            for arg in args:
                self.add(arg, **kwargs)
        elif len(args) > 0:
            if not isinstance(args[0], Datafile):
                raise InvalidInputException(
                    'Object "{}" must be of class Datafile to add it to a Dataset'.format(args[0])
                )
            self.files.add(args[0])

        else:
            # Add a single file, constructed by passing the arguments through to DataFile()
            self.files.add(Datafile(**kwargs))

    def get_file_by_label(self, label):
        """Get a single datafile from a dataset by filtering for files with the provided label.

        :param str label: the label to filter for
        :raise octue.exceptions.UnexpectedNumberOfResultsException: if zero or more than one results satisfy the filters
        :return octue.resources.datafile.DataFile:
        """
        return self.files.one(labels__contains=label)

    def download_all_files(self, local_directory=None):
        """Download all files in the dataset into the given local directory. If no path to a local directory is given,
        the files will be downloaded to temporary locations.

        :param str|None local_directory:
        :return None:
        """
        files_and_paths = []

        for datafile in self.files:
            if local_directory:
                path_relative_to_dataset = storage.path.relpath(datafile.cloud_path, self.path)
                local_path = os.path.abspath(os.path.join(local_directory, *path_relative_to_dataset.split("/")))
            else:
                local_path = None

            files_and_paths.append((datafile, local_path))

        def download(iterable_element):
            """Download a datafile to the given path.

            :param tuple(octue.resources.datafile.Datafile, str) iterable_element:
            :return str:
            """
            datafile = iterable_element[0]
            local_path = iterable_element[1]
            return datafile.download(local_path=local_path)

        # Use multiple threads to significantly speed up files downloads by reducing latency.
        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(download, files_and_paths)

    @staticmethod
    def _get_dataset_metadata(cloud_path):
        """Get the metadata for the given dataset if a dataset metadata file has previously been uploaded.

        :param str cloud_path: the path to the dataset cloud directory
        :return dict: the dataset metadata
        """
        storage_client = GoogleCloudStorageClient()
        metadata_file_path = storage.path.join(cloud_path, definitions.DATASET_METADATA_FILENAME)

        if not storage_client.exists(cloud_path=metadata_file_path):
            return {}

        return json.loads(storage_client.download_as_string(cloud_path=metadata_file_path))

    def _upload_dataset_metadata(self, cloud_path):
        """Upload a metadata file representing the dataset to the given cloud location.

        :param str cloud_path: the path to the dataset cloud directory
        :return None:
        """
        serialised_dataset = self.to_primitive()
        serialised_dataset["files"] = sorted(datafile.cloud_path for datafile in self.files)
        del serialised_dataset["path"]

        GoogleCloudStorageClient().upload_from_string(
            string=json.dumps(serialised_dataset),
            cloud_path=storage.path.join(cloud_path, definitions.DATASET_METADATA_FILENAME),
        )



import json
import os
import tempfile
from unittest.mock import patch
from google.auth.exceptions import DefaultCredentialsError
import google.oauth2.service_account

from octue.exceptions import InvalidInputException
from octue.cloud.credentials import GCPCredentialsManager
from tests.base import BaseTestCase


class TestGCPCredentialsManager(BaseTestCase):
    def test_error_raised_if_environment_variable_set_to_none(self):
        """Ensure an error is raised in the case that someone explicitly overrides the default value to None"""
        with self.assertRaises(InvalidInputException):
            GCPCredentialsManager(environment_variable_name=None)

    def test_error_raised_if_environment_variable_is_unset(self):
        """Ensure an error is raised if the given environment variable can't be found."""
        with self.assertRaises(InvalidInputException):
            mgr = GCPCredentialsManager(environment_variable_name="AN_UNSET_ENVVAR")

    def test_error_raised_if_environment_variable_is_invalid(self):
        """Ensure an error is raised if the given environment variable can't be found."""
        with patch.dict(os.environ, {"A_LEGIT_ENVVAR_NAME": "with-a-value-neither-file-nor-json"}):
            with self.assertRaises(InvalidInputException):
                mgr = GCPCredentialsManager(environment_variable_name="A_LEGIT_ENVVAR_NAME")
                mgr.get_credentials()

    def test_credentials_can_be_loaded_from_file(self):
        """Test that credentials can be loaded from the file at the path specified by the environment variable."""
        credentials = GCPCredentialsManager().get_credentials()
        self.assertEqual(type(credentials), google.oauth2.service_account.Credentials)

    def test_credentials_can_be_loaded_as_dictionary_from_string(self):
        """Test that credentials can be loaded as a dictionary from a string-valued environment variable."""
        with patch.dict(os.environ, {"GOOGLE_APPLICATION_CREDENTIALS": '{"blah": "nah"}'}):
            credentials = GCPCredentialsManager().get_credentials(as_dict=True)
            self.assertEqual(credentials, {"blah": "nah"})

    def test_credentials_can_be_loaded_as_dictionary_from_file(self):
        """Test that credentials can be loaded as a dictionary from a file."""
        with tempfile.NamedTemporaryFile(delete=False) as temporary_file:

            with open(temporary_file.name, "w") as f:
                json.dump({"blah": "nah"}, f)

            with patch.dict(os.environ, {"GOOGLE_APPLICATION_CREDENTIALS": temporary_file.name}):
                credentials = GCPCredentialsManager().get_credentials(as_dict=True)
                self.assertEqual(credentials, {"blah": "nah"})

    def test_default_credentials_are_used_when_environment_unset(self):
        """Test that application default credentials are accessed when the environment variable value is not set."""
        with patch.dict(os.environ, {"GOOGLE_APPLICATION_CREDENTIALS": ""}):
            with self.assertRaises(DefaultCredentialsError):
                GCPCredentialsManager().get_credentials()



from unittest import TestCase
from unittest.mock import Mock

from octue.utils.objects import get_nested_attribute, getattr_or_subscribe


class TestObjects(TestCase):
    def test_getattr_or_subscribe_with_dictionary(self):
        """Test that the getattr_or_subscribe function can get values from a dictionary."""
        self.assertEqual(getattr_or_subscribe(instance={"hello": "world"}, name="hello"), "world")

    def test_getattr_or_subscribe_with_object(self):
        """Test that the getattr_or_subscribe function can get attribute values from a class instance."""
        self.assertEqual(getattr_or_subscribe(instance=Mock(hello="world"), name="hello"), "world")

    def test_get_nested_attribute(self):
        """Test that nested attributes can be accessed."""
        inner_mock = Mock(b=3)
        outer_mock = Mock(a=inner_mock)
        self.assertEqual(get_nested_attribute(instance=outer_mock, nested_attribute_name="a.b"), 3)

    def test_get_nested_dictionary_attribute(self):
        """Test that nested attributes ending in a dictionary key can be accessed."""
        inner_mock = Mock(b={"hello": "world"})
        outer_mock = Mock(a=inner_mock)
        self.assertEqual(get_nested_attribute(instance=outer_mock, nested_attribute_name="a.b.hello"), "world")



from octue.mixins import Hashable
from octue.mixins.hashable import EMPTY_STRING_HASH_VALUE

from ..base import BaseTestCase


class EmptyClass:
    pass


class TypeWithNoAttributesToHash(Hashable):
    pass


class TypeWithAttributeToHash(Hashable):
    _ATTRIBUTES_TO_HASH = ("a",)
    a = 7


class TypeWithIterable(Hashable):
    _ATTRIBUTES_TO_HASH = ("my_iterable",)


class HashableTestCase(BaseTestCase):
    def test_no_attributes_to_hash_results_in_trivial_hash_value(self):
        """Assert classes with Hashable mixed in but no attributes to hash have a trivial hash value."""
        self.assertEqual(TypeWithNoAttributesToHash().hash_value, EMPTY_STRING_HASH_VALUE)

    def test_non_hashable_type_results_in_type_error(self):
        """Ensure trying to hash unhashable attributes results in a TypeError."""

        class MyClass(Hashable):
            _ATTRIBUTES_TO_HASH = ("a_class",)

        my_class = MyClass()
        my_class.a_class = EmptyClass()

        with self.assertRaises(TypeError):
            my_class.hash_value

    def test_iterable_attribute_with_mixed_types_raises_value_error(self):
        """Ensure trying to hash iterable attributes containing a mix between types with a 'hash_value' attribute and
        types that don't results in a TypeError.
        """
        my_class = TypeWithIterable()
        my_class.my_iterable = [1, 2, TypeWithNoAttributesToHash()]

        with self.assertRaises(ValueError):
            my_class.hash_value

    def test_unsortable_iterable_attribute_raises_type_error(self):
        """Ensure trying to hash unsortable iterable attributes results in a TypeError."""
        my_class = TypeWithIterable()
        my_class.my_iterable = [EmptyClass(), EmptyClass()]

        with self.assertRaises(TypeError):
            my_class.hash_value

    def test_set_hash_value(self):
        """Test that hash values can be set."""
        type_with_no_attributes_to_hash = TypeWithNoAttributesToHash()
        type_with_no_attributes_to_hash.hash_value = "hello"
        self.assertEqual(type_with_no_attributes_to_hash.hash_value, "hello")

    def test_set_hash_value_overrides_calculated_hash(self):
        """Test that hash values that are set override the calculated value."""
        type_with_hash = TypeWithAttributeToHash()
        self.assertEqual(len(type_with_hash.hash_value), 8)

        type_with_hash.hash_value = "hello"
        self.assertEqual(type_with_hash.hash_value, "hello")

    def test_reset_hash(self):
        """Test that hash values can be set and then reset to the calculated value."""
        type_with_hash = TypeWithAttributeToHash()
        original_calculated_hash = type_with_hash.hash_value
        self.assertEqual(len(original_calculated_hash), 8)

        type_with_hash.hash_value = "hello"
        self.assertEqual(type_with_hash.hash_value, "hello")

        type_with_hash.reset_hash()
        self.assertEqual(type_with_hash.hash_value, original_calculated_hash)

    def test_hash_value_cannot_be_set_if_hashable_has_immutable_hash_value(self):
        """Test that the hash value of a hashable instance with an immutable hash value cannot be set."""
        hashable = Hashable(immutable_hash_value="blue")

        with self.assertRaises(ValueError):
            hashable.hash_value = "red"



import uuid

from octue import exceptions
from octue.mixins import Identifiable

from ..base import BaseTestCase


class IdentifiableTestCase(BaseTestCase):
    def test_instantiates_with_no_args(self):
        """Ensures the class instantiates without arguments"""
        resource = Identifiable()
        self.assertIsInstance(resource.id, str)
        self.assertEqual(len(resource.id), 36)

    def test_instantiates_with_str_uuid(self):
        """Ensures class instantiates with a string uuid"""
        resource = Identifiable(id=str(uuid.uuid4()))
        self.assertIsInstance(resource.id, str)
        self.assertEqual(len(resource.id), 36)

    def test_instantiates_with_uuid(self):
        """Ensures class instantiates with a UUID()"""
        resource = Identifiable(id=uuid.uuid4())
        self.assertIsInstance(resource.id, str)
        self.assertEqual(len(resource.id), 36)

    def test_repr_with_no_name(self):
        """Test that the representation uses the ID if no name is provided."""
        resource = Identifiable(id="07d38e81-6b00-4079-901b-e250ea3c7773")
        self.assertEqual(repr(resource), f"<Identifiable({resource.id})>")

    def test_repr_with_name(self):
        """Test that the representation uses the name if one is provided."""
        resource = Identifiable(id="07d38e81-6b00-4079-901b-e250ea3c7773", name="my-name")
        self.assertEqual(repr(resource), "<Identifiable('my-name')>")

    def test_raises_error_with_non_uuid(self):
        """Ensures that if a string is passed not matching the UUID pattern, that an exception is raised"""
        with self.assertRaises(exceptions.InvalidInputException) as e:
            Identifiable(id="notauuid-6b00-4079-901b-e250ea3c7773")

        self.assertIn("is not a valid uuid string or instance of class UUID", e.exception.args[0])

    def test_raises_error_with_non_str_or_uuid(self):
        """Ensures that if an id is passed, it must be of type str or UUID"""

        class NotStrOrUUID:
            pass

        with self.assertRaises(exceptions.InvalidInputException) as e:
            Identifiable(id=NotStrOrUUID())

        self.assertIn("must be a valid uuid string, an instance of class UUID or None", e.exception.args[0])

    def test_str_on_subclass(self):
        """Ensures that calling `str` on an object inheriting from `Identifiable` will format it with the class name and
        ID.
        """

        class Inherit(Identifiable):
            pass

        resource = Inherit(id="07d38e81-6b00-4079-901b-e250ea3c7773")
        self.assertEqual(str(resource), "<Inherit(07d38e81-6b00-4079-901b-e250ea3c7773)>")

    def test_set_id_on_instantiated_object(self):
        """Ensures that setting id on an instantiated object will raise an error message, with a customised classname
        for inheritance
        """

        class Inherit(Identifiable):
            pass

        resource = Inherit()
        with self.assertRaises(AttributeError) as e:
            resource.id = "07d38e81-6b00-4079-901b-e250ea3c7773"

        self.assertIn("can't set attribute", e.exception.args[0])



import json
import os

import octue.migrations.manifest
from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.exceptions import InvalidInputException
from octue.migrations.cloud_storage import translate_bucket_name_and_path_in_bucket_to_cloud_path
from octue.mixins import Hashable, Identifiable, Pathable, Serialisable
from octue.resources.dataset import Dataset


class Manifest(Pathable, Serialisable, Identifiable, Hashable):
    """A representation of a manifest, which can contain multiple datasets This is used to manage all files coming into
    (or leaving), a data service for an analysis at the configuration, input or output stage.

    :param str|None id: the UUID of the manifest (a UUID is generated if one isn't given)
    :param str|None path: the path the manifest exists at (defaults to the current working directory)
    :param dict(str, octue.resources.dataset.Dataset|dict|str)|None datasets: a mapping of dataset names to `Dataset` instances, serialised datasets, or paths to datasets
    :return None:
    """

    _ATTRIBUTES_TO_HASH = ("datasets",)
    _SERIALISE_FIELDS = "datasets", "keys", "id", "name", "path"

    def __init__(self, id=None, path=None, datasets=None, **kwargs):
        if isinstance(datasets, list):
            datasets = octue.migrations.manifest.translate_datasets_list_to_dictionary(datasets, kwargs.get("keys"))

        super().__init__(id=id, path=path)
        self.datasets = {}

        # TODO The decoders aren't being used; utils.decoders.OctueJSONDecoder should be used in twined
        #  so that resources get automatically instantiated.
        #  Add a proper `decoder` argument  to the load_json utility in twined so that datasets, datafiles and manifests
        #  get initialised properly, then tidy up this hackjob. Also need to allow Pathables to update ownership
        #  (because decoders work from the bottom of the tree upwards, not top-down)

        self._instantiate_datasets(datasets or {})
        vars(self).update(**kwargs)

    @classmethod
    def from_cloud(
        cls,
        cloud_path=None,
        bucket_name=None,
        path_to_manifest_file=None,
    ):
        """Instantiate a Manifest from Google Cloud storage.

        :param str|None cloud_path: full path to manifest in cloud storage (e.g. `gs://bucket_name/path/to/manifest.json`)
        :return Dataset:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_to_manifest_file)

        bucket_name, path_to_manifest_file = storage.path.split_bucket_name_from_gs_path(cloud_path)

        serialised_manifest = json.loads(GoogleCloudStorageClient().download_as_string(cloud_path))

        datasets = {}

        for key, dataset in serialised_manifest["datasets"].items():
            datasets[key] = Dataset.from_cloud(dataset)

        return Manifest(
            id=serialised_manifest["id"],
            path=storage.path.generate_gs_path(bucket_name, path_to_manifest_file),
            datasets=datasets,
        )

    def to_cloud(self, cloud_path=None, bucket_name=None, path_to_manifest_file=None, store_datasets=False):
        """Upload a manifest to a cloud location, optionally uploading its datasets into the same directory.

        :param str|None cloud_path: full path to cloud storage location to store manifest at (e.g. `gs://bucket_name/path/to/manifest.json`)
        :param bool store_datasets: if True, upload datasets to same directory as manifest file
        :return str: gs:// path for manifest file
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_to_manifest_file)

        bucket_name, path_to_manifest_file = storage.path.split_bucket_name_from_gs_path(cloud_path)

        datasets = {}
        output_directory = storage.path.dirname(path_to_manifest_file)

        for key, dataset in self.datasets.items():
            if store_datasets:
                dataset_path = dataset.to_cloud(storage.path.generate_gs_path(bucket_name, output_directory))
                datasets[key] = dataset_path
            else:
                datasets[key] = dataset.cloud_path or dataset.absolute_path

        serialised_manifest = self.to_primitive()
        serialised_manifest["datasets"] = datasets
        del serialised_manifest["path"]

        GoogleCloudStorageClient().upload_from_string(string=json.dumps(serialised_manifest), cloud_path=cloud_path)
        return cloud_path

    @property
    def all_datasets_are_in_cloud(self):
        """Do all the files of all the datasets of the manifest exist in the cloud?

        :return bool:
        """
        return all(dataset.all_files_are_in_cloud for dataset in self.datasets.values())

    def get_dataset(self, key):
        """Get a dataset by its key (as defined in the twine).

        :param str key:
        :return octue.resources.dataset.Dataset:
        """
        dataset = self.datasets.get(key, None)

        if dataset is None:
            raise InvalidInputException(
                f"Attempted to fetch unknown dataset {key!r} from Manifest. Allowable keys are: "
                f"{list(self.datasets.keys())}"
            )

        return dataset

    def prepare(self, data):
        """Prepare new manifest from a manifest_spec.

        :param dict data:
        :return Manifest:
        """
        if len(self.datasets) > 0:
            raise InvalidInputException("You cannot `prepare()` a manifest already instantiated with datasets")

        for key, dataset_specification in data["datasets"].items():
            # TODO generate a unique name based on the filter key, label datasets so that the label filters in the spec
            #  apply automatically and generate a description of the dataset
            self.datasets[key] = Dataset(path_from=self, path=key)

        return self

    def _instantiate_datasets(self, datasets):
        """Add the given datasets to the manifest, instantiating them if needed and giving them the correct path.
        There are several possible forms the datasets can come in:
        * Instantiated Dataset instances
        * A list of dictionaries pointing to cloud datasets
        * Fully serialised form - includes path
        * `manifest.json` form - does not include path
        * Including datafiles that already exist
        * Including datafiles that don't yet exist or are not possessed currently (e.g. future output locations or
          cloud files)

        :param dict(str, octue.resources.dataset.Dataset|dict|str) datasets: the datasets to add to the manifest
        :return None:
        """
        for key, dataset in datasets.items():

            if isinstance(dataset, Dataset):
                self.datasets[key] = dataset

            else:
                # If `dataset` is just a path to a dataset:
                if isinstance(dataset, str):
                    if storage.path.is_qualified_cloud_path(dataset):
                        self.datasets[key] = Dataset.from_cloud(cloud_path=dataset, recursive=True)
                    else:
                        self.datasets[key] = Dataset.from_local_directory(path_to_directory=dataset, recursive=True)

                # If `dataset` is a dictionary including a "path" key:
                elif "path" in dataset:
                    # If the path is not a cloud path or an absolute local path:
                    if not os.path.isabs(dataset["path"]) and not storage.path.is_qualified_cloud_path(dataset["path"]):
                        path = dataset.pop("path")
                        self.datasets[key] = Dataset(**dataset, path=path, path_from=self)

                    # If the path is a cloud path or an absolute local path:
                    else:
                        self.datasets[key] = Dataset(**dataset)

                else:
                    self.datasets[key] = Dataset(**dataset, path=key, path_from=self)



import subprocess
from unittest.mock import Mock, patch

from octue.utils.processes import run_subprocess_and_log_stdout_and_stderr
from tests.base import BaseTestCase


class TestRunSubprocessAndLogStdoutAndStderr(BaseTestCase):
    def test_error_raised_if_process_fails(self):
        """Test that an error is raised if the subprocess fails."""
        mock_logger = Mock()

        with self.assertRaises(subprocess.CalledProcessError):
            with patch(
                "octue.utils.processes.Popen",
                return_value=MockPopen(stdout_messages=[b"bash: blah: command not found"], return_code=1),
            ):
                run_subprocess_and_log_stdout_and_stderr(command=["blah"], logger=mock_logger, shell=True)

        self.assertEqual(mock_logger.info.call_args[0][0], "bash: blah: command not found")

    def test_stdout_is_logged(self):
        """Test that any output to stdout from a subprocess is logged."""
        mock_logger = Mock()

        with patch(
            "octue.utils.processes.Popen", return_value=MockPopen(stdout_messages=[b"hello", b"goodbye"], return_code=0)
        ):
            process = run_subprocess_and_log_stdout_and_stderr(
                command=["echo hello && echo goodbye"], logger=mock_logger, shell=True
            )

        self.assertEqual(process.returncode, 0)
        self.assertEqual(mock_logger.info.call_args_list[0][0][0], "hello")
        self.assertEqual(mock_logger.info.call_args_list[1][0][0], "goodbye")


class MockPopen:
    """A mock of subprocess.Popen.

    :param iter(bytes) stdout_messages:
    :param int return_code:
    :return None:
    """

    def __init__(self, stdout_messages, return_code=0, *args, **kwargs):
        self.stdout = MockBufferedReader(stdout_messages)
        self.returncode = return_code

    def wait(self):
        """Do nothing.

        :return None:
        """
        pass


class MockBufferedReader:
    """A mock io.BufferedReader.

    :param iter(bytes) contents:
    :return None:
    """

    def __init__(self, contents):
        self.contents = contents
        self._line_count = 0

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def readline(self):
        """Read a line from the mocked buffered reader.

        :return str:
        """
        try:
            line = self.contents[self._line_count]
        except IndexError:
            raise StopIteration()

        self._line_count += 1
        return line
